import os

import requests


def prepare_uri(uri):
    if uri is None:
        return None

    if os.path.exists(uri):
        return uri

    prepared_uri = uri.replace("'", "")
    if prepared_uri:
        return prepared_uri
    else:
        return None


def download_file(url, target_path):
    response = requests.get(url)
    assert 199 < response.status_code < 300, f"Got status {response.status_code} for {url}"

    with open(target_path, "wb+") as fp:
        fp.write(response.content)

    return target_path



import argparse
import os
import shutil
import sys

import grapejuice_common.util
import grapejuice_common.variables as variables
from grapejuice_common import log_config
from grapejuice_common.dbus_client import dbus_connection
from grapejuice_common.gtk import gtk_stuff
from grapejuice_common.gtk.gtk_stuff import gtk_boot
from grapejuice_common.log_vacuum import vacuum_logs


def on_exit():
    p = variables.tmp_path()
    if os.path.exists(p) and os.path.isdir(p):
        shutil.rmtree(variables.tmp_path())


def main_gui():
    def make_main_window():
        from grapejuice_common import self_test
        self_test.post.run()

        from grapejuice.gui.main_window import MainWindow
        main_window = MainWindow()
        main_window.show()

    gtk_boot(make_main_window)


def func_gui(args):
    main_gui()


def func_post_install(args):
    from grapejuice_common import self_test
    self_test.post.run()

    from grapejuice import deployment
    deployment.post_install()


def func_player(args):
    def player_main():
        from grapejuice_common.settings import settings

        if settings.n_player_dialogs_remain > 0:
            settings.n_player_dialogs_remain = settings.n_player_dialogs_remain - 1
            settings.save()

            gtk_stuff.dialog("Currently, the Roblox Player is not confirmed to be working properly with Wine and "
                             "therefore is not supported by the Grapejuice project. Grapejuice will still launcher the "
                             "player, but it will most likely not work.")

        dbus_connection().play_game(grapejuice_common.util.prepare_uri(args.uri))

    gtk_boot(player_main, gtk_main=False)

    return 0


def func_studio(args):
    uri = grapejuice_common.util.prepare_uri(args.uri)
    if uri:
        is_local = False
        if not uri.startswith("roblox-studio:"):
            uri = "Z:" + uri.replace("/", "\\")
            is_local = True

        if is_local:
            dbus_connection().edit_local_game(uri)
        else:
            dbus_connection().edit_cloud_game(uri)

    else:
        dbus_connection().launch_studio()


def main(in_args=None):
    log_config.configure_logging("grapejuice")

    from grapejuice_common.settings import settings
    vacuum_logs()

    if settings:
        # TODO: Add logging for successful settings loading (Issue #9)
        pass

    if in_args is None:
        in_args = sys.argv

    parser = argparse.ArgumentParser(prog="grapejuice", description="Manage Roblox on Linux")
    subparsers = parser.add_subparsers(title="subcommands", help="sub-command help")

    parser_gui = subparsers.add_parser("gui")
    parser_gui.set_defaults(func=func_gui)

    parser_post_install = subparsers.add_parser("post_install")
    parser_post_install.set_defaults(func=func_post_install)

    parser_player = subparsers.add_parser("player")
    parser_player.add_argument("uri", type=str, help="Your Roblox token to join a game")
    parser_player.set_defaults(func=func_player)

    parser_studio = subparsers.add_parser("studio")
    parser_studio.add_argument(
        "uri",
        nargs="?",
        type=str,
        help="The URI or file to open roblox studio with",
        default=None
    )

    parser_studio.set_defaults(func=func_studio)

    args = parser.parse_args(in_args[1:])
    if hasattr(args, "func"):
        return args.func(args)
    else:
        parser.print_help()

    return 1


if __name__ == "__main__":
    main()



import os
import re
import time

import grapejuice_common.variables as variables
import grapejuice_common.winectrl as winectrl
from grapejuice_common.log_util import log_function
from grapejuice_common.util import download_file

DOWNLOAD_URL = "https://www.roblox.com/install/setup.ashx"


def set_graphics_mode(mode: int):
    settings_path = variables.wine_roblox_global_settings_13()
    if not os.path.exists(settings_path):
        return False

    ptn = r'token name=\"GraphicsMode\">(\d+)</token'
    gl = 'token name="GraphicsMode">' + str(mode) + '</token'

    output_lines = []
    with open(settings_path, "r") as fp:
        for line in fp.readlines():
            if "GraphicsMode" in line:
                line = re.sub(ptn, gl, line)

            output_lines.append(line)

    with open(settings_path, "w") as fp:
        fp.writelines(output_lines)


def set_graphics_mode_opengl():
    set_graphics_mode(4)


def get_installer():
    install_path = variables.installer_path()

    if os.path.exists(install_path):
        os.remove(install_path)

    download_file(DOWNLOAD_URL, install_path)


def run_installer():
    winectrl.create_prefix()
    get_installer()
    winectrl.run_exe_nowait(variables.installer_path())


@log_function
def locate_in_versions(exe_name):
    versions = os.path.join(variables.wine_roblox_prog(), "Versions")
    if not os.path.exists(versions):
        return None

    for dir in os.listdir(versions):
        fp = os.path.join(versions, dir)
        if os.path.isdir(fp):
            lp = os.path.join(fp, exe_name)
            if os.path.exists(lp) and os.path.isfile(lp):
                return lp


@log_function
def locate_roblox_exe(exe_name):
    versioned = locate_in_versions(exe_name)

    if not versioned:
        location = os.path.join(variables.wine_roblox_prog(), "Versions", exe_name)
        if os.path.exists(location):
            return location

    return versioned


@log_function
def locate_studio_launcher():
    return locate_roblox_exe("RobloxStudioLauncherBeta.exe")


@log_function
def locate_studio_exe():
    return locate_in_versions("RobloxStudioBeta.exe")


@log_function
def locate_player_launcher():
    return locate_in_versions("RobloxPlayerLauncher.exe")


@log_function
def locate_client_app_settings():
    studio_exe = locate_studio_exe()
    if studio_exe is None:
        return None

    return os.path.join(os.path.dirname(studio_exe), "ClientSettings", "ClientAppSettings.json")


def run_studio(uri="", ide=False):
    launcher = locate_studio_launcher()
    if launcher is None:
        return False

    if ide:
        winectrl.run_exe_nowait(launcher, "-ide", uri)
    else:
        if uri:
            winectrl.run_exe_nowait(launcher, uri)
        else:
            winectrl.run_exe_nowait(launcher, "-ide")

    return True


def studio_with_events(**events):
    studio_exe = locate_studio_exe()
    if studio_exe is None:
        return False

    args = [studio_exe]

    for k, v in events.items():
        args.append("-" + k)
        args.append(v)

    return winectrl.run_exe_nowait(*args)


def fast_flag_extract():
    fast_flag_path = variables.wine_roblox_studio_app_settings()
    if os.path.exists(fast_flag_path):
        os.remove(fast_flag_path)

    process = studio_with_events(startEvent="FFlagExtract", showEvent="NoSplashScreen")

    while True:
        if os.path.exists(fast_flag_path):
            stat = os.stat(fast_flag_path)
            if stat.st_size > 0:
                break

        time.sleep(0.5)

    process.kill()


def run_player(uri):
    player = locate_player_launcher()
    if player is None:
        return False

    winectrl.run_exe_nowait(player, uri)
    return True



import sys


def main(in_args):
    import gi
    gi.require_version("Gtk", "3.0")

    from gi.repository import Gtk
    from sparklepop.gui import MainWindow

    main_window = MainWindow()
    main_window.show()
    Gtk.main()


if __name__ == "__main__":
    main(sys.argv)



import logging
import os
from typing import List, Tuple


# Note about the self-test:
# Since part of the self-test is importing all project dependencies, Gtk gets imported.
# Gtk will cache sys.argv[0] and use this to display in the activities window.
# SO VERY IMPORTANT: MAKE SURE sys.arg[0] is hacked BEFORE Gtk is imported!

class TaskError(RuntimeError):
    def __init__(self, message: str, can_continue: bool = True):
        super().__init__(message)
        self._can_continue = can_continue

    @property
    def can_continue(self) -> bool:
        return self._can_continue


class TaskSequence:
    _ok_str = "Y"
    _fail_str = "N"
    _pending_str = "?"

    def __init__(self, name: str):
        self._name = name
        self._log = logging.getLogger(self._name)
        self._tasks: List[Tuple[str, callable]] = []
        self._results: List[str] = []
        self._reset_results()

    def task(self, name: str):
        tasks = self._tasks

        def wrap_function(func: callable):
            tasks.append((name, func))

            def wrapper():
                return func()

            return wrapper

        return wrap_function

    def _reset_results(self):
        self._results = ["?" for _ in self._tasks]

    def _log_results(self):
        self._log.info("Task status: " + "".join(self._results))

    def run(self):
        self._log.info("Running task sequence")
        self._reset_results()
        self._log_results()

        task_counter = 0
        for task_name, task_func in self._tasks:
            self._log.info(f"-- Running task -> {task_name} --")

            try:
                task_func(self._log)
                self._results[task_counter] = self._ok_str

            except TaskError as e:
                self._log.error(str(e))
                self._results[task_counter] = self._fail_str

                if e.can_continue:
                    self._log.warning(f"Task '{task_name}' failed, but we can continue")

                else:
                    self._log.error(f"Aborting due to failed task '{task_name}'")
                    raise e

            self._log.info(f"-- End of task --")

            task_counter += 1
            self._log_results()

        self._log.info(f"{self._name} OK")


post = TaskSequence("Power On Self Test")


@post.task("Read OS information")
def read_os_release(log: logging.Logger):
    os_release_path = "/etc/os-release"

    if not os.path.exists(os_release_path):
        raise TaskError(f"OS release information does not exist at '{os_release_path}'", can_continue=True)

    log.info(f"-- Path is {os_release_path}")
    with open(os_release_path, "r") as fp:
        for line in fp.readlines():
            log.info(line.strip())


@post.task("Check python dependencies")
def check_python_dependencies(log: logging.Logger):
    def import_psutil():
        import psutil

        return psutil

    def import_py_g_object():
        import gi
        gi.require_version("Gtk", "3.0")

        from gi.repository import Gtk
        from gi.repository import GObject
        from gi.repository import GLib

        return gi, Gtk, GObject, GLib

    def import_packaging():
        import packaging

        return packaging

    def import_requests():
        import requests

        return requests

    def import_wheel():
        import wheel

        return wheel

    def import_setuptools():
        import setuptools

        return setuptools

    def import_dbus():
        import dbus

        return dbus

    def try_import(import_func: callable):
        try:
            import_result = func()
            if isinstance(import_result, tuple):
                results = list(import_result)

            else:
                results = [import_result]

            log.info(f"Import function '{import_func.__name__}' loaded {len(results)} imports")

        except ImportError as e:
            msg = f"Import function '{import_func.__name__}' could not perform the imports it had to. " \
                  f"This installation of grapejuice is incorrect and should be rectified by doing a proper " \
                  f"installation. The underlying error is:\n {str(e)}"

            raise TaskError(msg, can_continue=False)

    map_23 = [
        import_psutil,
        import_py_g_object,
        import_packaging,
        import_requests,
        import_wheel,
        import_setuptools,
        import_dbus
    ]

    for func in map_23:
        try_import(func)



#!/usr/bin/env python3

import os
import shutil

import grapejuice.deployment as install
import grapejuice_common.variables as variables


def uninstall_desktop_files():
    for entry in os.listdir(variables.desktop_assets_dir()):
        file = os.path.join(variables.xdg_applications_dir(), entry)
        if os.path.exists(file) and os.path.isfile(file):
            os.remove(file)


def uninstall_mime_files():
    for mime in os.listdir(variables.mime_xml_assets_dir()):
        file = os.path.join(variables.xdg_mime_packages(), mime)

        if os.path.exists(file) and os.path.isfile(file):
            os.remove(file)


def remove_application_dir():
    shutil.rmtree(variables.system_application_dir(), ignore_errors=True)


def uninstall_main():
    uninstall_desktop_files()
    uninstall_mime_files()
    install.update_mime_database()
    remove_application_dir()


if __name__ == '__main__':
    uninstall_main()



import logging
import os
import time

from dbus import DBusException

import grapejuice_common.dbus_config as dbus_config
from grapejuice_common.pid_file import daemon_pid_file

LOG = logging.getLogger(__name__)


class DBusConnection:
    def __init__(self, connection_attempts=5, **kwargs):
        import dbus

        if "bus" in kwargs.keys():
            self.bus = kwargs["bus"]

        else:
            self.bus = dbus.SessionBus()

        self.pid_file = daemon_pid_file()
        self.daemon_alive = self.pid_file.is_running()
        self.proxy = None

        if not self.daemon_alive:
            LOG.debug("There is no Grapejuice daemon running, going to spawn one")
            self._spawn_daemon()
            self._wait_for_daemon(10)

        self._try_connect(attempts=connection_attempts)

    @property
    def connected(self):
        return self.proxy is not None

    def _wait_for_daemon(self, attempts: int):
        LOG.debug("Waiting for Grapejuice daemon to start")
        attempts_remaining = attempts

        while not self.daemon_alive and attempts_remaining > 0:
            attempts_remaining -= 1
            self.daemon_alive = self.pid_file.is_running(remove_junk=False)
            time.sleep(.5)

        if not self.daemon_alive:
            LOG.warning("Grapejuice daemon is not alive after waiting for it!")

    def _try_connect(self, attempts: int):
        attempts_remaining = attempts
        while attempts_remaining > 0 and not self.connected:
            LOG.debug(f"Connecting to Grapejuice daemon, attempt = {attempts - attempts_remaining}")

            attempts_remaining -= 1
            try:
                self.proxy = self.bus.get_object(dbus_config.bus_name, dbus_config.bus_path)

            except DBusException:
                self.daemon_alive = False

            if not self.connected:
                time.sleep(.5)

        if self.connected:
            LOG.debug("Connected to the Grapejuice daemon!")

    def launch_studio(self):
        return self.proxy.LaunchStudio()

    def play_game(self, uri):
        if uri:
            return self.proxy.PlayGame(uri)

        else:
            LOG.debug("No uri provided to play_game, returning False")

        return False

    def edit_local_game(self, place_path):
        return self.proxy.EditLocalGame(place_path)

    def edit_cloud_game(self, uri):
        if uri:
            return self.proxy.EditCloudGame(uri)

        return self.launch_studio()

    def install_roblox(self):
        self.proxy.InstallRoblox()

    def _spawn_daemon(self):
        LOG.debug("Spawning Grapejuice daemeon")
        os.spawnlp(os.P_NOWAIT, "python3", "python3", "-m", "grapejuiced", "daemonize")

    def version(self):
        return self.proxy.Version()

    def extract_fast_flags(self):
        self.proxy.ExtractFastFlags()

    def wine_version(self):
        wine_version = self.proxy.WineVersion()
        LOG.info(f"Wine version is {wine_version}")
        return wine_version


connection = None


def dbus_connection():
    global connection
    if connection is None:
        connection = DBusConnection()

    return connection



import os
from abc import ABC, abstractmethod

from grapejuice.update import legacy_update
from grapejuice_common.dist_info import DistributionType, dist_info

K_DISTRIBUTION_TYPE = "GRAPEJUICE_DISTRIBUTION_TYPE"


class UpdateProvider(ABC):
    @staticmethod
    @abstractmethod
    def can_update():
        raise NotImplementedError()

    @staticmethod
    def should_version_check():
        return False

    def do_update(self):
        pass

    def reinstall(self):
        pass


class SourceUpdateProvider(UpdateProvider):
    def can_update(self):
        return True

    def reinstall(self):
        legacy_update.update_and_reopen()

    def do_update(self):
        legacy_update.update_and_reopen()


class SystemPackageUpdateProvider(UpdateProvider):
    def can_update(self):
        return False


UPDATE_PROVIDER_MAPPING = {
    DistributionType.source: SourceUpdateProvider,
    DistributionType.system_package: SystemPackageUpdateProvider
}


def get_distribution_type():
    if K_DISTRIBUTION_TYPE in os.environ:
        return os.environ[K_DISTRIBUTION_TYPE]

    return dist_info.distribution_type


provider: UpdateProvider = UPDATE_PROVIDER_MAPPING[get_distribution_type()]()



import json
import logging
import os

from grapejuice_common import variables

LOG = logging.getLogger(__name__)


class DistributionType:
    source = "source"
    system_package = "system_package"


class DistributionInfo:
    def __init__(self, path: str = None):
        if path is None:
            path = os.path.join(variables.assets_dir(), "dist_info.json")

        self._path = path

        with open(self._path, "r") as fp:
            self._info = json.load(fp)

        LOG.debug(f"Loaded distribution info from {self._path}")
        LOG.info(f"Distribution type: {self.distribution_type}")

    def __getattr__(self, item):
        if item in self._info:
            return self._info[item]

        return super().__getattribute__(item)

    @property
    def distribution_type(self):
        k = "distribution_type"
        assert k in self._info

        v = self._info[k]
        assert isinstance(v, str)

        return v

    @distribution_type.setter
    def distribution_type(self, v: str):
        assert v in DistributionType.__dict__.values()
        self._info["distribution_type"] = v

    def write(self):
        with open(self._path, "w+") as fp:
            json.dump(self._info, fp)


dist_info = DistributionInfo()



import re

import requests
from packaging import version

from grapejuice_common import variables
from grapejuice_common.log_util import log_function

VERSION_PATTERN = re.compile(r'__version__\s*=\s*"(.+)?".*')
cached_remote_version = None


@log_function
def local_version():
    version_string = None

    with open(variables.src_init_py(), "r") as file:
        for line in file.readlines():
            match = VERSION_PATTERN.match(line)
            if match:
                version_string = match.group(1)

    if version_string is None:
        return version_string

    return version.parse(version_string)


@log_function
def remote_version():
    response = requests.get(variables.git_init_py_url())

    if 199 < response.status_code < 300:
        version_string = response.text.strip()

    else:
        return None

    match = VERSION_PATTERN.match(version_string)
    if match is None:
        return None

    global cached_remote_version
    cached_remote_version = version.parse(match.group(1))

    return cached_remote_version


@log_function
def update_available():
    local = local_version()
    if local is None:
        print("Warning: could not get the local version number")
        return False

    remote = remote_version()
    if remote is None:
        print("Warning: could not get the remote version number")
        return False

    return local < remote



import os
import subprocess

from grapejuice_common.errors import NoWineError

HERE = os.path.abspath(os.path.dirname(__file__))
K_GRAPEJUICE_INSTALL_PREFIX = "GRAPEJUICE_INSTALL_PREFIX"
K_GRAPEJUICE_IS_PACKAGING = "GRAPEJUICE_IS_PACKAGING"
K_GRAPEJUICE_PACKAGE_PREFIX = "GRAPEJUICE_PACKAGE_PREFIX"


def is_packaging() -> bool:
    return K_GRAPEJUICE_IS_PACKAGING in os.environ and os.environ[K_GRAPEJUICE_IS_PACKAGING].lower() == "yes"


def ensure_dir(p):
    if not os.path.exists(p):
        os.makedirs(p)

    return p


def home():
    return os.environ["HOME"]


def system_application_dir():
    p = os.path.dirname(src_dir())
    assert os.path.exists(p)
    return p


def user_application_dir():
    return os.path.join(local_share(), "grapejuice")


def packaging_prefix():
    if is_packaging():
        return os.environ[K_GRAPEJUICE_PACKAGE_PREFIX]

    return installation_prefix()


def assets_dir():
    search_locations = [
        os.path.join(HERE, "assets"),
        os.path.join(os.getcwd(), "assets"),
        os.path.join(system_application_dir(), "assets")
    ]

    for p in search_locations:
        if os.path.exists(p):
            return p

    raise RuntimeError("Could not find assets directory")


def desktop_assets_dir():
    return os.path.join(assets_dir(), "desktop")


def mime_xml_assets_dir():
    return os.path.join(assets_dir(), "mime_xml")


def icons_assets_dir():
    return os.path.join(assets_dir(), "icons")


def installation_prefix():
    if K_GRAPEJUICE_INSTALL_PREFIX in os.environ:
        return os.environ[K_GRAPEJUICE_INSTALL_PREFIX]

    return os.path.dirname(system_application_dir())


def src_dir():
    return os.path.abspath(os.path.dirname(HERE))


def glade_dir():
    return os.path.join(assets_dir(), "glade")


def grapejuice_glade():
    return os.path.join(glade_dir(), "grapejuice.glade")


def about_glade():
    return os.path.join(glade_dir(), "about.glade")


def fast_flag_editor_glade():
    return os.path.join(glade_dir(), "fast_flag_editor.glade")


def grapejuice_components_glade():
    return os.path.join(glade_dir(), "grapejuice_components.glade")


def fast_flag_warning_glade():
    return os.path.join(glade_dir(), "fast_flag_warning.glade")


def sparklepop_glade():
    return os.path.join(glade_dir(), "sparklepop.glade")


def config_base_dir():
    return ensure_dir(os.path.join(xdg_config_home(), "brinkervii"))


def sparklepop_config_dir():
    return ensure_dir(os.path.join(config_base_dir(), "sparklepop"))


def grapejuice_config_dir():
    return ensure_dir(os.path.join(config_base_dir(), "grapejuice"))


def grapejuice_user_settings():
    return os.path.join(grapejuice_config_dir(), "user_settings.json")


def sparklepop_snapshots_dir():
    return ensure_dir(os.path.join(sparklepop_config_dir(), "snapshots"))


def src_init_py():
    search_locations = [
        os.path.join(src_dir(), "grapejuice", "__init__.py"),
        os.path.join(src_dir(), "__init__.py"),
        os.path.join(src_dir(), "src", "grapejuice", "__init__.py")
    ]

    for p in search_locations:
        if os.path.exists(p):
            return p


def wineprefix_dir():
    return os.path.join(user_application_dir(), "wineprefix")


def wine_drive_c():
    return os.path.join(wineprefix_dir(), "drive_c")


def wine_roblox_prog():
    return os.path.join(wine_drive_c(), "Program Files (x86)", "Roblox")


def wine_temp():
    return ensure_dir(os.path.join(wine_drive_c(), "windows", "temp"))


def wine_user():
    return os.path.join(wine_drive_c(), "users", os.environ["USER"])


def wine_roblox_appdata():
    return os.path.join(wine_user(), "Local Settings", "Application Data", "Roblox")


def wine_roblox_global_settings_13():
    return os.path.join(wine_roblox_appdata(), "GlobalSettings_13.xml")


def wine_roblox_studio_app_settings():
    return os.path.join(wine_roblox_appdata(), "ClientSettings", "StudioAppSettings.json")


def installer_path():
    return os.path.join(wine_temp(), "Roblox_Installer.exe")


def xdg_applications_dir():
    return os.path.join(local_share(), "applications")


def xdg_mime_dir():
    return os.path.join(local_share(), "mime")


def xdg_mime_packages():
    return os.path.join(xdg_mime_dir(), "packages")


def xdg_icons():
    return os.path.join(local_share(), "icons")


def xdg_config_home():
    if "XDG_CONFIG_HOME" in os.environ:
        config_home = os.environ["XDG_CONFIG_HOME"]
        if config_home and os.path.exists(config_home) and os.path.isdir(config_home):
            return config_home

    config_home = os.path.join(home(), ".config")
    if not os.path.exists(config_home):
        os.makedirs(config_home)

    return config_home


def local_share():
    return os.path.join(dot_local(), "share")


def local_var():
    return os.path.join(dot_local(), "var")


def local_log():
    return os.path.join(local_var(), "log")


def logging_directory():
    return os.path.join(local_log(), "grapejuice")


def dot_local():
    if is_packaging():
        return installation_prefix()

    path = os.path.join(home(), ".local")
    if not os.path.exists(path):
        os.makedirs(path)

    return path


def xdg_documents():
    run = subprocess.run(["xdg-user-dir", "DOCUMENTS"], stdout=subprocess.PIPE)
    documents_path = run.stdout.decode("utf-8").rstrip()

    if os.path.exists(documents_path):
        return documents_path

    documents_path = os.path.join(home(), "Documents")
    return ensure_dir(documents_path)


def git_repository():
    return "https://gitlab.com/brinkervii/grapejuice"


def git_wiki():
    return git_repository() + "/-/wikis/home"


def git_init_py_url():
    return git_repository() + "/raw/master/src/grapejuice/__init__.py"


def git_zip_download():
    return git_repository() + "/-/archive/master/grapejuice-master.zip"


def tmp_path():
    d = "grapejuice-{}".format(os.getpid())
    p = os.path.join("/tmp", d)

    if not os.path.exists(p):
        os.makedirs(p)
    else:
        if not os.path.isdir(p):
            os.remove(p)
            os.makedirs(p)

    return p


def tmp_zip_path():
    return os.path.join(tmp_path(), "grapejuice-download.zip")


def wine_binary(arch=""):
    path_search = []
    if "PATH" in os.environ:
        for spec in os.environ["PATH"].split(":"):
            path_search.append(os.path.join(spec, "wine" + arch))

    static_search = [
        "/opt/wine-stable/bin/wine" + arch,
        "/opt/wine-staging/bin/wine" + arch
    ]

    for p in (path_search + static_search):
        if os.path.exists(p):
            return p

    raise NoWineError()


def wine_binary_64():
    return wine_binary("64")


def required_wine_version():
    return "wine-4.0"



import os
import shutil
import sys
import zipfile

import grapejuice_common.variables as variables
from grapejuice_common.pid_file import daemon_pid_file
from grapejuice_common.util import download_file


def perform_download():
    download_path = variables.tmp_zip_path()
    if os.path.exists(download_path):
        os.remove(download_path)

    return download_file(
        variables.git_zip_download(),
        download_path
    )


def unpack_download(filename):
    p = variables.tmp_path()
    try:
        with zipfile.ZipFile(filename, 'r') as zip_ref:
            zip_ref.extractall(p)

    except zipfile.BadZipFile as e:
        print(repr(e))
        return None

    return p


def perform_update():
    daemon_pid = daemon_pid_file()
    if daemon_pid.is_running():
        daemon_pid.kill()

    filename = perform_download()
    if os.path.exists(filename):
        download_dir = unpack_download(filename)
        if download_dir is None:
            return False

        src_dir = os.path.join(download_dir, "grapejuice-master")
        if os.path.exists(src_dir):
            os.chdir(src_dir)
            os.spawnlp(os.P_WAIT, "python3", "python3", "./install.py")
            return True

        else:
            return False
    else:
        return False


def delete_tmp():
    p = variables.tmp_path()
    if os.path.exists(p) and os.path.isdir(p):
        shutil.rmtree(p)


def update_and_reopen():
    src_dir = variables.src_dir()

    if perform_update():
        os.chdir(src_dir)
        delete_tmp()
        os.spawnlp(os.P_NOWAIT, "python3", "python3", "-m", "grapejuice", "gui")
        sys.exit(0)



import os

from grapejuice import background
from grapejuice.tasks import DisableMimeAssociations, ApplyDLLOverrides, InstallRoblox, DeployAssociations, \
    GraphicsModeOpenGL, SandboxWine, RunRobloxStudio, ExtractFastFlags, OpenLogsDirectory
from grapejuice.update.update_provider import provider as update_provider
from grapejuice_common import variables, robloxctrl
from grapejuice_common import winectrl, version
from grapejuice_common.errors import NoWineError
from grapejuice_common.event import Event
from grapejuice_common.gtk.gtk_stuff import WindowBase, dialog
from grapejuice_common.settings import settings
from grapejuice_common.winectrl import wine_ok

on_destroy = Event()

once_task_tracker = dict()


def on_task_removed(task: background.BackgroundTask):
    if task in once_task_tracker.keys():
        once_task_tracker[task] = None


background.tasks.task_removed.add_listener(on_task_removed)


def run_task_once(task_class, on_already_running: callable, *args, **kwargs):
    if task_class in once_task_tracker.values():
        on_already_running()
        return

    task = task_class(*args, **kwargs)
    once_task_tracker[task] = task_class

    background.tasks.add(task)


def generic_already_running():
    dialog("This task is already being performed!")


def xdg_open(*args):
    os.spawnlp(os.P_NOWAIT, "xdg-open", "xdg-open", *args)


class MainWindowHandlers:
    def on_destroy(self, *_):
        from gi.repository import Gtk
        on_destroy()
        Gtk.main_quit()

    def run_winecfg(self, *_):
        winectrl.winecfg()

    def run_regedit(self, *_):
        winectrl.regedit()

    def run_winetricks(self, *_):
        winectrl.wine_tricks()

    def disable_mime_assoc(self, *_):
        run_task_once(DisableMimeAssociations, generic_already_running)

    def sandbox(self, *_):
        run_task_once(SandboxWine, generic_already_running)

    def run_roblox_installer(self, *_):
        def no_wine_dialog() -> None:
            dialog("Grapejuice could not find a working Wine binary, please install Wine using your operating "
                   "system's package manager in order to install and use Roblox.")

            return None

        try:
            wine_bin = variables.wine_binary()
            if not os.path.exists(wine_bin):
                return no_wine_dialog()

        except NoWineError:
            return no_wine_dialog()

        if not wine_ok():
            return

        run_task_once(InstallRoblox, generic_already_running)

    def run_roblox_studio(self, *_):
        studio_launcher_location = robloxctrl.locate_studio_launcher()
        if not studio_launcher_location:
            dialog("Grapejuice could not locate Roblox Studio. You might have to install it first by going to the "
                   "maintenance tab and clicking 'Install Roblox'")
            return

        if not wine_ok():
            return

        run_task_once(RunRobloxStudio, generic_already_running)

    def wine_explorer(self, *_):
        winectrl.explorer()

    def apply_dll_overrides(self, *_):
        run_task_once(ApplyDLLOverrides, generic_already_running)

    def open_drive_c(self, *_):
        xdg_open(variables.wine_drive_c())

    def show_about(self, *_):
        from grapejuice.gui.about_window import AboutWindow
        wnd = AboutWindow()
        wnd.window.run()

        del wnd

    def open_fast_flag_editor(self, *_):
        def open_editor(b):
            if not b:
                return

            task = ExtractFastFlags()

            def poll():
                if task.finished:
                    from grapejuice.gui.fast_flag_editor import FastFlagEditor
                    wnd = FastFlagEditor()
                    wnd.window.show()

                return not task.finished

            from gi.repository import GObject
            GObject.timeout_add(100, poll)

            background.tasks.add(task)

        if settings.show_fast_flag_warning:
            from grapejuice.gui.fast_flag_warning import FastFlagWarning
            wnd = FastFlagWarning(open_editor)
            wnd.show()

        else:
            open_editor(True)

    def show_wiki(self, *_):
        xdg_open(variables.git_wiki())

    def perform_update(self, *_):
        dialog("If the Grapejuice upgrade breaks your installation, please redo the Grapejuice installation according "
               "to the instructions in the Grapejuice git repository. The upgrade will begin after you close this "
               "dialog.")

        update_provider.do_update()

    def reinstall(self, *_):
        update_provider.reinstall()

    def deploy_assocs(self, *_):
        run_task_once(DeployAssociations, generic_already_running)

    def launch_sparklepop(self, *_):
        os.spawnlp(os.P_NOWAIT, "python", "python", "-m", "sparklepop")

    def graphicsmode_opengl(self, *_):
        run_task_once(GraphicsModeOpenGL, generic_already_running)

    def open_logs_directory(self, *_):
        run_task_once(OpenLogsDirectory, generic_already_running)


class MainWindow(WindowBase):
    def __init__(self):
        super().__init__(
            variables.grapejuice_glade(),
            MainWindowHandlers()
        )

        self.update_status_label.set_text("Checking for updates...")
        self.update_update_related_buttons()
        self.update_update_status()

        background.tasks.tasks_changed.add_listener(self.on_tasks_changed)
        on_destroy.add_listener(self.before_destroy)

        self.on_tasks_changed()

    @property
    def deploy_associations_button(self):
        return self.builder.get_object("deploy_associations_button")

    @property
    def reinstall_button(self):
        return self.builder.get_object("reinstall_button")

    @property
    def update_status_label(self):
        return self.builder.get_object("update_status_label")

    @property
    def update_button(self):
        return self.builder.get_object("update_button")

    def update_update_related_buttons(self):
        if not update_provider.can_update():
            self.deploy_associations_button.hide()
            self.reinstall_button.hide()

    def update_update_status(self):
        w = self

        class CheckUpdates(background.BackgroundTask):
            def __init__(self):
                super().__init__("Checking for a newer version of Grapejuice")

            def run(self) -> None:
                if update_provider.can_update():
                    if version.update_available():
                        s = "This version of Grapejuice is out of date\n{} -> {}".format(
                            str(version.local_version()),
                            str(version.cached_remote_version)
                        )

                        w.update_status_label.set_text(s)
                        w.update_button.show()
                    else:
                        local_ver = version.local_version()
                        if local_ver > version.cached_remote_version:
                            s = "This version of Grapejuice is from the future\n{}".format(str(local_ver))
                        else:
                            s = "Grapejuice is up to date\n{}".format(str(local_ver))

                        w.update_status_label.set_text(s)
                        w.update_button.hide()

                else:
                    s = f"Running Grapejuice {version.local_version()}"
                    w.update_status_label.set_text(s)
                    w.update_button.hide()

                self.finish()

        background.tasks.add(CheckUpdates())

    @property
    def window(self):
        return self.builder.get_object("main_window")

    @property
    def background_task_spinner(self):
        return self.builder.get_object("background_task_spinner")

    def on_tasks_changed(self):
        if background.tasks.count > 0:
            self.background_task_spinner.start()

        else:
            self.background_task_spinner.stop()

    def show(self):
        self.window.show()

    def before_destroy(self):
        background.tasks.tasks_changed.remove_listener(self.on_tasks_changed)



import glob
import os
import shutil
import subprocess
from typing import Dict

from grapejuice_common import variables
from grapejuice_common.pid_file import daemon_pid_file
from grapejuice_common.settings import settings

K_GRAPEJUICE_EXECUTABLE = "GRAPEJUICE_EXECUTABLE"

DESKTOP_STUDIO = "roblox-studio.desktop"
DESKTOP_PLAYER = "roblox-player.desktop"

PROTOCOL_ASSOC = {
    "x-scheme-handler/roblox-studio": DESKTOP_STUDIO,
    "x-scheme-handler/roblox-player": DESKTOP_PLAYER
}

FILE_ASSOC = {
    "application/x-roblox-rbxl": DESKTOP_STUDIO,
    "application/x-roblox-rbxlx": DESKTOP_STUDIO
}


def log(*args):
    print("!!", *args)


def invoke_xdg_mime() -> bool:
    return not variables.is_packaging()


def get_desktop_install_path(desktop_name):
    return os.path.join(variables.xdg_applications_dir(), desktop_name)


def install_desktop_files(environ: Dict):
    assert isinstance(environ, dict)

    desktop_assets = variables.desktop_assets_dir()

    for desktop_name in os.listdir(desktop_assets):
        target_path = get_desktop_install_path(desktop_name)
        source_path = os.path.join(desktop_assets, desktop_name)

        target_directory = os.path.dirname(target_path)
        os.makedirs(target_directory, exist_ok=True)

        log("Installing desktop file", source_path, "to", target_path)

        def process_line(line):
            for k, v in environ.items():
                search_ptn = "$" + k
                if search_ptn in line:
                    line = line.replace(search_ptn, v)

            return line

        with open(source_path, "r") as source_fp:
            with open(target_path, "w+") as target_fp:
                output_lines = list(map(process_line, source_fp.readlines()))
                target_fp.writelines(output_lines)


def update_mime_database():
    if not invoke_xdg_mime():
        return

    log("Updating MIME database")
    xdg_mime = variables.xdg_mime_dir()

    os.makedirs(xdg_mime, exist_ok=True)
    subprocess.check_call(["update-mime-database", xdg_mime])


def install_mime_def(src, target_dir):
    log("Installing MIME definition", src, "into the directory", target_dir)
    name = os.path.basename(src)

    target = os.path.join(target_dir, name)
    assert target

    if os.path.exists(target):
        os.remove(target)

    shutil.copy(src, target)


def install_mime_files():
    mime_source = variables.mime_xml_assets_dir()
    mime_packages = variables.xdg_mime_packages()

    log("Installing MIME XML files from", mime_source, "into", mime_packages)

    os.makedirs(mime_packages, exist_ok=True)

    for file in glob.glob(os.path.join(mime_source, "*")):
        install_mime_def(file, mime_packages)

    update_mime_database()


def mime_assoc(desktop, mime_type):
    if not invoke_xdg_mime():
        return

    log("MIME assoc", desktop, mime_type)

    subprocess.check_call(["xdg-mime", "default", desktop, mime_type])


def update_protocol_handlers():
    log("Updating protocol handlers")

    for scheme, desktop in PROTOCOL_ASSOC.items():
        mime_assoc(desktop, scheme)


def update_file_associations():
    log("Updating file associations")

    for scheme, desktop in FILE_ASSOC.items():
        mime_assoc(desktop, scheme)


def install_icons():
    icons = glob.glob(os.path.join(variables.icons_assets_dir(), "**"), recursive=True)
    icons = list(filter(os.path.isfile, icons))

    common_prefix = variables.icons_assets_dir()

    rel_icons = list(map(lambda path: os.path.relpath(path, common_prefix), icons))

    for icon in rel_icons:
        src = os.path.join(common_prefix, icon)
        dst = os.path.join(variables.xdg_icons(), icon)
        log("Installing icon", src, dst)

        dst_parent = os.path.dirname(dst)
        os.makedirs(dst_parent, exist_ok=True)

        shutil.copy(src, dst)


def post_install():
    assert variables.K_GRAPEJUICE_INSTALL_PREFIX in os.environ

    def application_dir():
        if variables.is_packaging():
            return os.path.join(variables.packaging_prefix(), "share", "grapejuice")

        return variables.system_application_dir()

    def grapejuice_executable():
        if K_GRAPEJUICE_EXECUTABLE in os.environ:
            return os.environ[K_GRAPEJUICE_EXECUTABLE]

        return os.path.join(application_dir(), "bin", "grapejuice")

    environ = {
        variables.K_GRAPEJUICE_INSTALL_PREFIX: variables.installation_prefix(),
        "APPLICATION_DIR": application_dir(),
        K_GRAPEJUICE_EXECUTABLE: grapejuice_executable(),
        "GRAPEJUICE_ICON": "grapejuice",
        "STUDIO_ICON": "grapejuice-roblox-studio",
        "PLAYER_ICON": "grapejuice-roblox-player"
    }

    pid_file = daemon_pid_file()
    pid_file.kill()

    install_mime_files()
    install_icons()
    install_desktop_files(environ)
    update_protocol_handlers()
    update_file_associations()

    settings.performed_post_install = True
    settings.save()



import atexit
import json
import logging
import os
import time
from typing import Union

from grapejuice_common import variables

LOG = logging.getLogger(__name__)


class Setting:
    def __init__(self, key: str, display_name: str, default_value: Union[any, None] = None):
        self._key: str = key
        self._display_name: str = display_name
        self._default_value: Union[any, None] = default_value
        self._value: any = default_value

    def isinstance(self, v):
        return isinstance(self.value, v)

    def reset(self):
        self._value = self._default_value

    @property
    def key(self) -> str:
        return self._key

    @property
    def value(self) -> Union[any, None]:
        return self._value

    @value.setter
    def value(self, v) -> None:
        if self._default_value is not None:
            assert type(self._default_value) is type(v)

        self._value = v

    @property
    def display_name(self) -> str:
        return self._display_name

    def __repr__(self) -> str:
        return f"'{self.key}' -> '{self.value}' ({self._default_value})"


class UserSettings:
    def __init__(self, file_location=variables.grapejuice_user_settings()):
        self._location = file_location

        self.performed_post_install = False
        self.n_player_dialogs_remain = 3
        self.show_fast_flag_warning = True

        self.wine_binary = Setting("wine_binary", "Wine binary", "")
        self.ignore_wine_version = Setting("ignore_wine_version", "Ignore Wine version", False)

        self._update_last_run()

        self.load()

    def _filtered_dict(self):
        d = dict()

        for k, v in self.__dict__.items():
            if isinstance(v, Setting):
                d[v.key] = v.value

            elif not k.startswith("_"):
                d[k] = v

        return d

    def _update_last_run(self):
        self.last_run = int(time.time() // 1)
        LOG.debug(f"Set last run time to {self.last_run}")

    def _accept_json(self, o):
        self_dict = self.__dict__

        LOG.debug("Loading JSON object into the settings file")

        for k, v in o.items():
            LOG.debug(f"Read '{k}' with value '{v}'")

            if k in self_dict:
                LOG.debug(f"Going to apply the value with key {k}")
                existing_value = self_dict[k]

                if isinstance(existing_value, Setting):
                    LOG.debug(f"Doing a special setting thing with {k}")
                    existing_value.value = v
                    continue

                else:
                    LOG.debug(f"Applying {k}")
                    self_dict[k] = v

            else:
                LOG.debug(f"Not going to apply a value with key {k}, because it is not on the settings object")

    def load(self):
        if os.path.exists(self._location):
            LOG.debug(f"Loading settings from '{self._location}'")

            with open(self._location, "r") as fp:
                json_object = json.load(fp)
                self._accept_json(json_object)

        else:
            LOG.debug("There is no settings file present, going to save one")
            self.save()

    def save(self):
        self._update_last_run()

        LOG.debug(f"Saving settings file to '{self._location}'")
        with open(self._location, "w+") as fp:
            json.dump(self._filtered_dict(), fp)

    def ui_facing_settings(self):
        return list(filter(lambda v: isinstance(v, Setting), self.__dict__.values()))


settings = UserSettings()


def save_settings():
    LOG.info("Saving settings...")
    settings.save()


atexit.register(save_settings)



import re


class RegistryType:
    def __init__(self, type_string):
        self.type_string = type_string

    def __repr__(self):
        return self.type_string

    def __hash__(self):
        return hash(self.type_string)


class UndefinedRegistryType(RegistryType):
    def __init__(self):
        super().__init__("undefined")


class DWORDRegistryType(RegistryType):
    def __init__(self):
        super().__init__("dword")


class StringRegistryType(RegistryType):
    def __init__(self):
        super().__init__("string")


class HexRegistryType(RegistryType):
    def __init__(self):
        super().__init__("hex")


def get_registry_value_type(value):
    dword_match = re.match(r"dword:(.+)", value)
    if dword_match:
        return DWORDRegistryType(), dword_match.group(1)

    hex_match = re.match(r"hex:(.+)", value)
    if hex_match:
        return HexRegistryType(), hex_match.group(1)

    string_match = re.match(r"\"(.*)?\"", value)
    if string_match:
        return StringRegistryType(), string_match.group(1)

    return UndefinedRegistryType(), value


registry_item_counter = 0


class RegistryItem:
    def __init__(self):
        global registry_item_counter
        self.serial_no = registry_item_counter
        registry_item_counter += 1

    def __hash__(self):
        return hash(self.serial_no)

    def __lt__(self, other):
        return self.serial_no < other.serial_no


class RegistryKey(RegistryItem):
    def __init__(self, key, timestamp):
        super().__init__()
        self.key = str(key)
        self.timestamp = int(timestamp)
        self.tags = []
        self.properties = []
        self.comments = []
        self.text_items = []
        self.ats = []

    def __repr__(self):
        return self.key

    def __hash__(self):
        my_hash = hash(super())
        my_hash ^= hash(self.key)
        my_hash ^= hash(self.timestamp)

        def hash_list(l, h):
            for item in l:
                h ^= hash(item)

            return h

        my_hash = hash_list(self.tags, my_hash)
        my_hash = hash_list(self.properties, my_hash)
        my_hash = hash_list(self.comments, my_hash)
        my_hash = hash_list(self.text_items, my_hash)
        my_hash = hash_list(self.ats, my_hash)

        return my_hash

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __ne__(self, other):
        return not self.__eq__(other)


class RegistryTag(RegistryItem):
    def __init__(self, key, value):
        super().__init__()
        self.key = key
        self.value = value

    def __repr__(self):
        return "#" + self.key + "=" + self.value

    def __hash__(self):
        return hash(super()) ^ hash(self.key) ^ hash(self.value)


class RegistryAt(RegistryItem):
    def __init__(self, value):
        super().__init__()
        self.value = value

    def __hash__(self):
        return hash(super()) ^ hash(self.value)


class RegistryProperty(RegistryItem):
    def __init__(self, key, value):
        super().__init__()
        self.key = key
        self.type, self.value = get_registry_value_type(value)

    def __repr__(self):
        return self.key + " = " + repr(self.value)

    def __hash__(self):
        return hash(super()) ^ hash(self.key) ^ hash(self.value) ^ hash(self.type)


class RegistryComment(RegistryItem):
    def __init__(self, text):
        super().__init__()
        self.text = text

    def __repr__(self):
        return ";; " + self.text

    def __hash__(self):
        return hash(super()) ^ hash(self.text)


class RegistryText(RegistryItem):
    def __init__(self, text):
        super().__init__()
        self.text = text

    def __repr__(self):
        return self.text

    def __hash__(self):
        return hash(super()) ^ hash(self.text)


class RegFile:
    def __init__(self, p):
        self.global_key = RegistryKey("", 0)
        self.keys: [RegistryKey] = []
        read_reg(self, p)

    def to_dict(self):
        d = dict()
        for key in self.keys:
            d[key.key] = key

        return d


def read_reg(rf: RegFile, p: str):
    global registry_item_counter
    registry_item_counter = 0

    current_key = None

    with open(p) as file:
        for line in file.readlines():
            if not line:
                continue

            line = line.strip("\n")
            if not line:
                continue

            key_match = re.match(r"\[(.+)?]\s*(\d+)", line)
            if key_match:
                current_key = RegistryKey(key_match.group(1), key_match.group(2))
                rf.keys.append(current_key)
                continue

            property_match = re.match(r"\"(.+)?\"=(.+)", line)
            if property_match:
                if current_key:
                    prop = RegistryProperty(property_match.group(1), property_match.group(2))
                    current_key.properties.append(prop)
                continue

            tag_match = re.match(r"#(.+)?=(.*)", line)
            if tag_match:
                tag = RegistryTag(tag_match.group(1), tag_match.group(2))
                if current_key:
                    current_key.tags.append(tag)
                else:
                    rf.global_key.tags.append(tag)
                continue

            comment_match = re.match(r";;\s*(.*)", line)
            if comment_match:
                comment = RegistryComment(comment_match.group(1))
                if current_key:
                    current_key.comments.append(comment)
                else:
                    rf.global_key.comments.append(comment)
                continue

            at_match = re.match(r"@=(.+)", line)
            if at_match:
                at = RegistryAt(at_match.group(1))
                if current_key:
                    current_key.ats.append(at)
                else:
                    rf.global_key.ats.append(at)

            text = RegistryText(line)
            if current_key:
                current_key.text_items.append(text)
            else:
                rf.global_key.text_items.append(text)



from abc import ABC, abstractmethod

from gi.repository import Gtk

from grapejuice_common import variables


class GrapejuiceComponent(ABC):
    def __init__(self):
        self._builder = Gtk.Builder()
        self._builder.add_from_file(variables.grapejuice_components_glade())

    @abstractmethod
    def get_root_widget(self):
        pass



import logging
import threading
from typing import Union

from grapejuice_common.event import Event


class Task:
    _log: logging.Logger

    def __init__(self, name):
        self._log = logging.getLogger(f"{Task.__name__}/{name}")

        self._name = name
        self._finished = False
        self._collection: Union[None, TaskCollection] = None

    @property
    def finished(self):
        return self._finished

    @property
    def collection(self):
        return self._collection

    @collection.setter
    def collection(self, value):
        assert self._collection is None, "Can only set collection once"
        self._collection = value

    def on_finished(self):
        pass

    def finish(self):
        self.on_finished()
        self._finished = True

        if self.collection is not None:
            self.collection.remove(self)

    @property
    def name(self):
        return self._name


class BackgroundTask(threading.Thread, Task):
    def __init__(self, name="Untitled task"):
        threading.Thread.__init__(self)
        Task.__init__(self, name)

    def run(self) -> None:
        raise NotImplementedError()

    def __repr__(self):
        return "BackgroundTask: {}".format(self._name)


class MockBackgroundTask(Task):
    def __init__(self, name):
        super().__init__(name)

    def start(self):
        pass


class TaskCollection:
    def __init__(self):
        self._tasks = []

        self.task_added = Event()
        self.task_removed = Event()
        self.tasks_changed = Event()

    def add(self, task: BackgroundTask):
        task.collection = self
        self._tasks.append(task)
        task.start()

        self.task_added(task)
        self.tasks_changed()

    def remove(self, task):
        self._tasks.remove(task)
        self.task_removed(task)
        self.tasks_changed()

    @property
    def count(self):
        return len(self._tasks)


tasks = TaskCollection()



import logging
import os
import re
import shutil
import signal
import subprocess
import sys
import time
from subprocess import DEVNULL
from typing import List

import grapejuice_common.variables as variables
from grapejuice_common.log_util import log_on_call, log_function

LOG = logging.getLogger(__name__)

space_version_ptn = re.compile(r"wine-(.+?)\s+")
non_space_version_ptn = re.compile(r"wine-(.+)")
space = " "


class ProcessWrapper:
    def __init__(self, proc: subprocess.Popen):
        self.proc = proc

    @property
    def exited(self):
        proc = self.proc

        if proc.returncode is None:
            proc.poll()

        return proc.returncode is not None

    def kill(self):
        if not self.exited:
            os.kill(self.proc.pid, signal.SIGINT)

    def __del__(self):
        del self.proc


processes: List[ProcessWrapper] = []
is_polling = False


@log_on_call("Preparing for Wine")
def prepare():
    prefix_dir = variables.wineprefix_dir()
    os.environ["WINEPREFIX"] = prefix_dir
    os.environ["WINEARCH"] = "win64"

    if not os.path.exists(prefix_dir):
        os.makedirs(prefix_dir)


@log_on_call("Running Wine configuration")
def winecfg():
    prepare()
    os.spawnlp(os.P_NOWAIT, variables.wine_binary(), variables.wine_binary(), "winecfg")


@log_on_call("Running registry editor")
def regedit():
    prepare()
    os.spawnlp(os.P_NOWAIT, variables.wine_binary(), variables.wine_binary(), "regedit")


@log_on_call("Running Windows Explorer")
def explorer():
    prepare()
    os.spawnlp(os.P_NOWAIT, variables.wine_binary(), variables.wine_binary(), "explorer")


def load_reg(srcfile):
    LOG.info(f"Loading registry file {srcfile} into the wineprefix")
    prepare()
    target_filename = str(int(time.time())) + ".reg"
    target_path = os.path.join(variables.wine_temp(), target_filename)
    shutil.copyfile(srcfile, target_path)

    winreg = "C:\\windows\\temp\\{}".format(target_filename)
    os.spawnlp(os.P_WAIT, variables.wine_binary(), variables.wine_binary(), "regedit", "/S", winreg)
    os.spawnlp(os.P_WAIT, variables.wine_binary_64(), variables.wine_binary_64(), "regedit", "/S", winreg)

    os.remove(target_path)


def load_regs(s: [str], patches: dict = None):
    prepare()
    target_filename = str(int(time.time())) + ".reg"
    target_path = os.path.join(variables.wine_temp(), target_filename)

    with open(target_path, "w+") as fp:
        if patches is None:
            fp.write("\r\n".join(s))
        else:
            out_lines = []
            for line in s:
                for k, v in patches.items():
                    varkey = "$" + k
                    if varkey in line:
                        line = line.replace(varkey, v)

                out_lines.append(line)

            fp.writelines(out_lines)

    winreg = "C:\\windows\\temp\\{}".format(target_filename)
    os.spawnlp(os.P_WAIT, variables.wine_binary(), variables.wine_binary(), "regedit", "/S", winreg)
    os.spawnlp(os.P_WAIT, variables.wine_binary_64(), variables.wine_binary_64(), "regedit", "/S", winreg)

    os.remove(target_path)


@log_on_call("Running Winetricks")
def wine_tricks():
    prepare()
    os.spawnlp(os.P_NOWAIT, "winetricks", "winetricks")


@log_on_call("Disabling MIME associations in wineprefix")
def disable_mime_assoc():
    load_reg(os.path.join(variables.assets_dir(), "disable_mime_assoc.reg"))


def set_roblox_document_path():
    src_path = os.path.join(variables.assets_dir(), "roblox_documents_folder.reg")
    patches = dict()

    documents_dir = "Z:" + variables.xdg_documents().replace("/", "\\\\")
    patches["DOCUMENTS_DIR"] = documents_dir
    LOG.info(f"Setting the roblox documents directory to '{documents_dir}'")

    with open(src_path, "r") as fp:
        load_regs(fp.readlines(), patches)


@log_on_call("Loading DLL overrides")
def load_dll_overrides():
    load_reg(os.path.join(variables.assets_dir(), "dll_overrides.reg"))


@log_on_call("Sandboxing user directories in the wineprefix")
def sandbox():
    user_dir = variables.wine_user()

    if os.path.exists(user_dir) and os.path.isdir(user_dir):
        for file in os.listdir(user_dir):
            p = os.path.join(user_dir, file)

            if os.path.islink(p):
                LOG.info(f"Sandboxing {file}")
                os.remove(p)
                os.makedirs(p, exist_ok=True)


@log_on_call("Configuring the wineprefix")
def configure_prefix():
    disable_mime_assoc()
    sandbox()
    load_dll_overrides()
    set_roblox_document_path()


@log_on_call("Creating the wineprefix")
def create_prefix():
    configure_prefix()


@log_function
def prefix_exists():
    return os.path.exists(variables.wineprefix_dir())


@log_function
def run_exe(exe_path, *args):
    prepare()
    if len(args) > 0:
        os.spawnlp(os.P_NOWAIT, variables.wine_binary(), variables.wine_binary(), exe_path, *args)
    else:
        os.spawnlp(os.P_NOWAIT, variables.wine_binary(), variables.wine_binary(), exe_path)


@log_function
def run_exe_nowait(exe_path, *args) -> ProcessWrapper:
    prepare()

    command = [variables.wine_binary(), exe_path, *args]
    p = subprocess.Popen(command, stdin=DEVNULL, stdout=sys.stdout, stderr=sys.stderr, close_fds=True)

    wrapper = ProcessWrapper(p)
    processes.append(wrapper)

    poll_processes()

    return wrapper


def _poll_processes() -> bool:
    """
    Makes sure zombie launchers are taken care of
    :return: Whether or not processes remain
    """
    global is_polling
    exited = []

    for proc in processes:
        if proc.exited:
            exited.append(proc)

    for proc in exited:
        processes.remove(proc)
        del proc

    processes_left = len(processes) > 0
    if not processes_left:
        is_polling = False

    return processes_left


def poll_processes():
    global is_polling
    if is_polling:
        return

    from gi.repository import GObject
    GObject.timeout_add(100, _poll_processes)


@log_function
def wine_ok(system_wine: str = None, show_dialog=True):
    from grapejuice_common.settings import settings
    if settings.ignore_wine_version.value:
        return True

    from grapejuice_common.dbus_client import dbus_connection
    from grapejuice_common.gtk.gtk_stuff import dialog

    def prepare_version(s):
        from packaging import version

        if space in s:
            match = space_version_ptn.match(s)

        else:
            match = non_space_version_ptn.match(s)

        assert match is not None

        return version.parse(match.group(1))

    if system_wine is None:
        system_wine = prepare_version(dbus_connection().wine_version())

    else:
        system_wine = prepare_version(system_wine)

    required_wine = prepare_version(variables.required_wine_version())

    def version_to_string(v):
        if v.public:
            return v.public

        if v.base_version:
            return v.base_version

        return repr(v)

    if system_wine < required_wine:
        if show_dialog:
            msg = "Your system has Wine version {} installed, you need at least Wine version {} in order to run Roblox." \
                .format(version_to_string(system_wine), version_to_string(required_wine))

            dialog(msg)

        return False

    return True



import math

from grapejuice_common.event import Event


class Paginator:
    def __init__(self, collection, page_size):
        self._collection = collection
        self._page_size = page_size
        self._current_page = 0
        self._filter_function = None

        self.paged = Event()

    @property
    def _filtered_collection(self):
        collection = self._collection

        if callable(self._filter_function):
            collection = list(self._filter_function(self._collection))

        if hasattr(collection, "sort"):
            collection.sort()

        return collection

    @property
    def page(self):
        coll = self._filtered_collection

        lower_limit = self._current_page * self._page_size
        upper_limit = min(len(coll), lower_limit + self._page_size)

        return coll[lower_limit:upper_limit]

    @property
    def current_page_index(self):
        return self._current_page

    @property
    def n_pages(self):
        return math.ceil(len(self._filtered_collection) / self._page_size)

    @property
    def filter_function(self):
        return self._filter_function

    @filter_function.setter
    def filter_function(self, v):
        assert v is None or callable(v)

        self._filter_function = v
        self._current_page = max(0, min(self.n_pages - 1, self._current_page))
        self.paged()

    @property
    def at_first_page(self):
        return self._current_page == 0

    @property
    def at_last_page(self):
        return self._current_page >= self.n_pages - 1

    def next(self):
        if not self.at_last_page:
            self._current_page = min(self.n_pages - 1, self._current_page + 1)
            self.paged()

    def previous(self):
        if not self.at_first_page:
            self._current_page = max(0, self._current_page - 1)
            self.paged()



from gi.repository import Gtk


class SparklepopHandlers:
    parent = None
    flag = True

    def on_destroy(self, *args):
        Gtk.main_quit()

    def sig_debug(self, *args):
        if self.flag:
            self.parent.hide_info_panel()
        else:
            self.parent.show_info_panel()

        self.flag = not self.flag

    def snapshot_info_changed(self, *args):
        self.parent.on_snapshot_changed()

    def snapshot_info_changed(self, *args):
        self.parent.on_snapshot_changed()

    def create_snapshot(self, *args):
        import grapejuice_registry.snapshot as snapshot
        snapshot.create_snapshot()
        self.parent.populate_snapshots()



from grapejuice_common import WindowBase
from grapejuice_common import variables
from grapejuice_registry.snapshot import Snapshot
from sparklepop.gui.snapshot_view_model import SnapshotViewModel


class SnapshotWidget:
    def __init__(self, parent, snap: Snapshot):
        from gi.repository import Gtk

        self.parent = parent
        self.snap = snap
        self.container = Gtk.Button(always_show_image=False)
        self.container.set_focus_on_click(False)

        self.box = Gtk.Box()
        self.box.set_orientation(1)
        sides = ("top", "left", "bottom", "right")
        for side in sides:
            getattr(self.box, "set_margin_" + side)(10)

        self.container.add(self.box)

        self.name_label = Gtk.Label(label=snap.name)
        self.name_label.set_xalign(0)
        self.box.add(self.name_label)

        self.datetime_label = Gtk.Label(label=snap.datetime)
        self.datetime_label.set_xalign(0)
        self.datetime_label.set_margin_bottom(5)
        self.box.add(self.datetime_label)

        self.description_label = Gtk.Label(label=snap.description)
        self.description_label.set_xalign(0)
        self.description_label.set_line_wrap(True)
        self.box.add(self.description_label)

        self.bad_indicator = Gtk.Image()
        self.bad_indicator.set_halign(Gtk.Align.END)
        self.bad_indicator.set_margin_top(5)
        self.displayed_broken = snap.broken
        self.box.add(self.bad_indicator)

        self.container.connect("clicked", self._on_click)

    def _on_click(self, *args):
        self.parent.select_snapshot(self)

    @property
    def displayed_name(self):
        return self.name_label.get_text()

    @displayed_name.setter
    def displayed_name(self, value):
        self.name_label.set_text(value)

    @property
    def displayed_broken(self):
        return self.snap.broken

    @displayed_broken.setter
    def displayed_broken(self, value):
        from gi.repository import Gtk
        icon = "gtk-dialog-error" if value else "gtk-yes"
        self.bad_indicator.set_from_stock(icon, Gtk.IconSize.BUTTON)

    @property
    def displayed_description(self):
        return self.description_label.get_text()

    @displayed_description.setter
    def displayed_description(self, value):
        self.description_label.set_text(value)


class MainWindow(WindowBase):
    snapshot_model = None
    selected_snap_widget: SnapshotWidget = None
    setting_snap_widget = False
    made_changes = False

    def __init__(self):
        from sparklepop.gui import SparklepopHandlers

        def create_handlers():
            handlers = SparklepopHandlers()
            handlers.parent = self
            return handlers

        super().__init__(variables.sparklepop_glade(), create_handlers)
        self._build()
        self.hide_info_panel()

    def _snapshot_box(self):
        return self.builder.get_object("snapshot_box")

    def _snapshot_info_panel(self):
        return self.builder.get_object("snapshot_info_panel")

    def _snapshot_info_id(self):
        return self.builder.get_object("snapshot_info_id")

    def _snapshot_info_name(self):
        return self.builder.get_object("snapshot_info_name")

    def _snapshot_info_broken(self):
        return self.builder.get_object("snapshot_info_broken")

    def _snapshot_info_description(self):
        return self.builder.get_object("snapshot_info_description")

    def _window(self):
        return self.builder.get_object("sparklepop_main")

    def hide_info_panel(self):
        self._snapshot_info_panel().hide()

    def show_info_panel(self):
        self._snapshot_info_panel().show()

    def show(self):
        self.populate_snapshots()
        self._window().show()

    def populate_snapshots(self):
        from gi.repository import Gtk
        snapshot_box = self._snapshot_box()

        for child in snapshot_box.get_children():
            snapshot_box.remove(child)

        first_pfx = True
        self.snapshot_model = SnapshotViewModel()
        for pfx in self.snapshot_model:
            pfx_label = Gtk.Label(use_markup=True)
            pfx_label.set_markup("<span size='large' weight='bold'>Prefix {}</span>".format(pfx.id))
            pfx_label.set_xalign(0)
            pfx_label.set_selectable(False)
            pfx_label.set_margin_top(0 if first_pfx else 20)
            snapshot_box.add(pfx_label)

            pfx_container = Gtk.FlowBox(homogeneous=True)

            for snap in pfx:
                widget = SnapshotWidget(self, snap)
                pfx_container.add(widget.container)

            first_pfx = False
            snapshot_box.add(pfx_container)

        snapshot_box.show_all()

    def select_snapshot(self, widget: SnapshotWidget):
        self.setting_snap_widget = True

        from gi.repository import Gtk
        if self.selected_snap_widget is not None:
            self.save_snap()

        self.selected_snap_widget = widget
        if widget is not None:
            snap = widget.snap
            self._snapshot_info_id().set_text(snap.id)
            self._snapshot_info_name().set_text(snap.name)
            self._snapshot_info_broken().set_active(snap.broken)

            buf = Gtk.TextBuffer()
            buf.set_text(snap.description)
            self._snapshot_info_description().set_buffer(buf)
            self.show_info_panel()

        self.made_changes = False
        self.setting_snap_widget = False

    def _info_name(self):
        return self._snapshot_info_name().get_text()

    def _info_broken(self):
        return self._snapshot_info_broken().get_active()

    def _info_description(self):
        buf = self._snapshot_info_description().get_buffer()
        return buf.get_text(buf.get_start_iter(), buf.get_end_iter(), False)

    def on_snapshot_changed(self):
        if self.setting_snap_widget or self.selected_snap_widget is None:
            return

        self.selected_snap_widget.displayed_name = self._info_name()
        self.selected_snap_widget.displayed_broken = self._info_broken()
        self.selected_snap_widget.displayed_description = self._info_description()

        self.made_changes = True

    def save_snap(self):
        if self.selected_snap_widget is None or not self.made_changes:
            return

        snap = self.selected_snap_widget.snap
        snap.name = self._info_name()
        snap.broken = self._info_broken()
        snap.description = self._info_description()
        snap.update()



import json
import os
import re
import time
import uuid
from datetime import datetime
from zipfile import ZipFile

from grapejuice_common import variables

cached_snapshot_index = None


class Snapshot:
    loaded = False

    def __init__(self, pfx_id, snapshot_id=None):
        if snapshot_id is None:
            snapshot_id = str(uuid.uuid4())

        self.metadata = {
            "id": snapshot_id,
            "name": "Untitled snapshot",
            "description": "Snapshot created at " + str(datetime.now()),
            "pfx_id": pfx_id,
            "timestamp": int(time.time()),
            "broken": False
        }

    def __lt__(self, other):
        if isinstance(other, Snapshot):
            return self.timestamp > other.timestamp

        return False

    @property
    def id(self):
        return self.metadata["id"]

    @property
    def filename(self):
        return self.id + ".zip"

    @property
    def metadata_filename(self):
        return self.id + ".metadata.json"

    @property
    def name(self):
        return self.metadata["name"] or "Untitled snapshot"

    @name.setter
    def name(self, value):
        self.metadata["name"] = str(value)

    @property
    def description(self):
        return self.metadata["description"] or "The description for this snapshot was lost"

    @description.setter
    def description(self, value):
        self.metadata["description"] = str(value)

    @property
    def pfx_id(self):
        return self.metadata["pfx_id"] or "undefined"

    @pfx_id.setter
    def pfx_id(self, value):
        self.metadata["pfx_id"] = value

    @property
    def timestamp(self):
        return self.metadata["timestamp"] or int(time.time())

    @property
    def datetime(self):
        return datetime.utcfromtimestamp(self.timestamp).strftime("%Y-%m-%d %H:%M:%S")

    @property
    def broken(self):
        value = self.metadata["broken"]
        if value is None:
            value = False

        return value

    @broken.setter
    def broken(self, value):
        self.metadata["broken"] = value

    def create(self, source_list: [(str, str)]):
        p_zip = os.path.join(variables.sparklepop_snapshots_dir(), self.filename)
        with ZipFile(p_zip, "w") as z:
            for name, p_file in source_list:
                z.write(p_file, name)

        self.save_metadata()

    def _metadata_path(self):
        return os.path.join(variables.sparklepop_snapshots_dir(), self.metadata_filename)

    def save_metadata(self):
        with open(self._metadata_path(), "w+") as f_metadata:
            self.metadata["archive"] = self.filename
            pretty_json_dump(self.metadata, f_metadata)

        if not self.loaded:
            snapshot_index_append(self.pfx_id, self.id)

    def load(self):
        with open(self._metadata_path(), "r") as f_metadata:
            self.loaded = True
            self.metadata = json.load(f_metadata)

    def update(self):
        self.save_metadata()


def pretty_json_dump(obj: dict, fp):
    json.dump(obj, fp, sort_keys=True, indent=4)


def generate_pfx_metadata():
    return {
        "id": str(uuid.uuid4())
    }


def fill_metadata_gaps(target: dict, source: dict) -> bool:
    modified = False

    for k in source.keys():
        if k in target.keys():
            if type(source[k]) != type(target[k]):
                del target[k]

    for k in source.keys():
        if k not in target.keys():
            target[k] = source[k]
            modified = True

    return modified


def current_pfx_metadata() -> dict:
    p = os.path.join(variables.wineprefix_dir(), "sparklepop.metadata.json")
    if not os.path.exists(p):
        metadata = generate_pfx_metadata()
        with open(p, "w+") as file:
            json.dump(metadata, file)
    else:
        with open(p, "r") as file:
            metadata = json.load(file)

    if fill_metadata_gaps(metadata, generate_pfx_metadata()):
        with open(p, "w+") as file:
            json.dump(metadata, file)

    return metadata


def snapshot_index_path():
    return os.path.join(variables.sparklepop_config_dir(), "snapshot_index.json")


def snapshot_index():
    global cached_snapshot_index
    if cached_snapshot_index is not None:
        return cached_snapshot_index

    p_index = snapshot_index_path()
    if os.path.exists(p_index):
        with open(p_index, "r") as index_file:
            index = json.load(index_file)
    else:
        index = {
            "last_updated": int(time.time()),
            "snapshots": {}
        }

        with open(p_index, "w+") as index_file:
            pretty_json_dump(index, index_file)

    cached_snapshot_index = index
    return index


def save_snapshot_index(index):
    index["last_updated"] = int(time.time())
    with open(snapshot_index_path(), "w+") as file:
        pretty_json_dump(index, file)

    global cached_snapshot_index
    cached_snapshot_index = index


def snapshot_index_append(pfx_id, snap_id):
    index = snapshot_index()

    if pfx_id in index["snapshots"].keys():
        snap_object = index["snapshots"][pfx_id]
    else:
        snap_object = {}
        index["snapshots"][pfx_id] = snap_object

    if "snapshots" in snap_object.keys():
        snap_array = snap_object["snapshots"]
    else:
        snap_array = []
        snap_object["snapshots"] = snap_array

    snap_array.append(snap_id)
    snap_object["last_updated"] = int(time.time())
    save_snapshot_index(index)


def create_snapshot():
    reg_files = []
    pfx_basepath = variables.wineprefix_dir()
    for file in os.listdir(pfx_basepath):
        if re.match(r"^.+?\.reg", file):
            reg_files.append((file, os.path.join(pfx_basepath, file)))

    snapshot = Snapshot(current_pfx_metadata()["id"])
    snapshot.create(reg_files)



from grapejuice_registry.reg_file import RegFile, RegistryKey


class RegDiff:
    def __init__(self, baseline: RegFile, other: RegFile):
        self.baseline: RegFile = baseline
        self.other: RegFile = other

        self.additions: [RegistryKey] = []
        self.removals: [RegistryKey] = []
        self.changes: [(RegistryKey, RegistryKey)] = []

    def do_diff(self):
        baseline = self.baseline.to_dict()
        other = self.other.to_dict()
        all_keys = set()
        add_keys = []
        removal_keys = []
        change_keys = []

        for key, reg_key in baseline.items():
            if key not in other.keys():
                removal_keys.append(key)
                all_keys.add(key)

        for key, reg_key in other.items():
            if key not in baseline.keys():
                add_keys.append(key)
                all_keys.add(key)

        for key in removal_keys:
            all_keys.remove(key)

        for key in add_keys:
            all_keys.remove(key)

        for key in all_keys:
            if baseline[key] != other[key]:
                change_keys.append(key)

        for key in add_keys:
            self.additions.append(other[key])

        for key in removal_keys:
            self.removals.append(baseline[key])

        for key in change_keys:
            self.changes.append((baseline[key], other[key]))


if __name__ == "__main__":
    import os

    broken = os.path.join(os.environ["HOME"], ".local/share/grapejuice", "wineprefix-broken/user.reg")
    working = os.path.join(os.environ["HOME"], ".local/share/grapejuice", "wineprefix-working/user.reg")

    diff = RegDiff(RegFile(broken), RegFile(working))
    diff.do_diff()
    debug = 0



from grapejuice_common.gtk.grapejuice_component import GrapejuiceComponent
from grapejuice_common.paginator import Paginator


class GtkPaginator(GrapejuiceComponent):
    def __init__(self, paginator: Paginator):
        super().__init__()
        self._paginator = paginator

        def go_back(*_):
            self._paginator.previous()

        def go_forward(*_):
            self._paginator.next()

        def update_display():
            self._label.set_text(self._label_text)

        self._button_previous.connect("clicked", go_back)
        self._button_next.connect("clicked", go_forward)
        self._paginator.paged.add_listener(update_display)

        update_display()

    @property
    def _label(self):
        return self._builder.get_object("paginator_label")

    @property
    def _label_text(self):
        return "{}/{}".format(self._paginator.current_page_index + 1, self._paginator.n_pages)

    @property
    def _button_previous(self):
        return self._builder.get_object("paginator_previous")

    @property
    def _button_next(self):
        return self._builder.get_object("paginator_next")

    def get_root_widget(self):
        return self._builder.get_object("paginator")



import os
import subprocess

from dbus import DBusException

from grapejuice import background, deployment
from grapejuice_common import winectrl, robloxctrl, variables
from grapejuice_common.dist_info import dist_info, DistributionType


def install_roblox():
    from grapejuice_common.dbus_client import dbus_connection
    dbus_connection().install_roblox()


class DisableMimeAssociations(background.BackgroundTask):
    def __init__(self):
        super().__init__("Disabling Wine associations")

    def run(self) -> None:
        winectrl.disable_mime_assoc()
        self.finish()


class ApplyDLLOverrides(background.BackgroundTask):
    def __init__(self):
        super().__init__("Applying DLL overrides")

    def run(self) -> None:
        winectrl.load_dll_overrides()
        self.finish()


class InstallRoblox(background.BackgroundTask):
    def __init__(self):
        super().__init__("Installing Roblox")

    def run(self) -> None:
        try:
            install_roblox()
        except DBusException:
            pass  # TODO: find a proper fix

        self.finish()


class DeployAssociations(background.BackgroundTask):
    def __init__(self):
        super().__init__("Deploying associations")

    def run(self) -> None:
        if dist_info.distribution_type == DistributionType.source:
            os.environ[variables.K_GRAPEJUICE_INSTALL_PREFIX] = variables.dot_local()
            deployment.post_install()

        else:
            self._log.error(f"{DeployAssociations.__name__} is only supported on source distributions")

        self.finish()


class GraphicsModeOpenGL(background.BackgroundTask):
    def __init__(self):
        super().__init__("Changing the Roblox GraphicsMode to OpenGL")

    def run(self) -> None:
        robloxctrl.set_graphics_mode_opengl()
        self.finish()


class SandboxWine(background.BackgroundTask):
    def __init__(self):
        super().__init__("Sandboxing the Wine prefix")

    def run(self) -> None:
        winectrl.sandbox()
        self.finish()


class RunRobloxStudio(background.BackgroundTask):
    def __init__(self):
        super().__init__("Launching Roblox Studio")

    def run(self) -> None:
        from grapejuice_common.dbus_client import dbus_connection
        dbus_connection().launch_studio()
        self.finish()


class ExtractFastFlags(background.BackgroundTask):
    def __init__(self):
        super().__init__("Extracting Fast Flags")

    def run(self) -> None:
        try:
            from grapejuice_common.dbus_client import dbus_connection
            dbus_connection().extract_fast_flags()

        except DBusException:
            pass

        self.finish()


class OpenLogsDirectory(background.BackgroundTask):
    def __init__(self):
        super().__init__("Opening logs directory")

    def run(self) -> None:
        path = variables.logging_directory()
        os.makedirs(path, exist_ok=True)

        subprocess.check_call(["xdg-open", path])
        self.finish()



from grapejuice_common import variables
from grapejuice_common.gtk.gtk_stuff import WindowBase
from grapejuice_common.settings import settings


class FastFlagWarning(WindowBase):
    def __init__(self, callback):
        super().__init__(variables.fast_flag_warning_glade(), self)

        self._do_continue = False
        self._callback = callback

        self.devforum_link.set_label("Read more on the Roblox Developer forum")
        self.warn_check.set_active(settings.show_fast_flag_warning)

    @property
    def window(self):
        return self.builder.get_object("fast_flag_warning")

    @property
    def devforum_link(self):
        return self.builder.get_object("devforum_link")

    @property
    def warn_check(self):
        return self.builder.get_object("warn_check")

    def destroy(self):
        self.window.destroy()

    def on_close(self, *_):
        if self._do_continue:
            settings.show_fast_flag_warning = self.warn_check.get_active()
            settings.save()

        self._callback(self._do_continue)

    def abort(self, *_):
        self._do_continue = False
        self.destroy()

    def open_editor(self, *_):
        self._do_continue = True
        self.destroy()

    def show_forum_post(self, link_button):
        pass

    def show(self):
        self.window.show()



import os
import sys

here = os.path.abspath(os.path.dirname(__file__))
project_path = here
src_path = os.path.join(project_path, "src")
readme_path = os.path.join(project_path, "README.md")
requirements_path = os.path.join(project_path, "requirements.txt")

sys.path.insert(0, src_path)

from setuptools import setup, find_packages, Command

import grapejuice_packaging.metadata as metadata

with open(readme_path, "r") as fp:
    long_description = fp.read()

with open(requirements_path, "r") as fp:
    requirements = [r.lstrip().rstrip() for r in fp.readlines()]


class PackageDebian(Command):
    description = "Package Grapejuice for debian"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        from grapejuice_packaging.__main__ import main as packaging_main
        packaging_main([sys.argv[0], "debian"])


setup(
    name="grapejuice",
    author=metadata.author_name,
    author_email=metadata.author_email,
    version=metadata.package_version,
    description=metadata.package_description,
    license=metadata.package_license,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=metadata.package_repository,
    classifiers=[
        "Development Status :: 4 - Beta",
        'Programming Language :: Python :: 3.7'
    ],
    keywords=["grapejuice wine roblox studio"],
    packages=find_packages("src", exclude=("grapejuice_packaging",)),
    package_dir={"": "src"},
    include_package_data=True,
    python_requires=">=3.7",
    install_requires=requirements,
    entry_points={
        "console_scripts": [
            "grapejuice=grapejuice.__main__:main",
            "grapejuiced=grapejuiced.__main__:main"
        ]
    },
    cmdclass={
        "package_debian": PackageDebian
    }
)



from grapejuice_registry import snapshot


class PrefixModel(list):
    def __init__(self, id: str, timestamp: int):
        super().__init__()
        self.id = id
        self.timestamp = timestamp

    def __lt__(self, other):
        if isinstance(other, PrefixModel):
            return self.timestamp > other.timestamp

        return False


class SnapshotViewModel(list):
    def __init__(self):
        super().__init__()

        index = snapshot.snapshot_index()
        snapshots_table = index["snapshots"]
        for pfx_id, pfx_object in snapshots_table.items():
            pfx = PrefixModel(pfx_id, pfx_object["last_updated"])

            for snap_id in pfx_object["snapshots"]:
                snap = snapshot.Snapshot(pfx_id, snap_id)
                snap.load()
                pfx.append(snap)

            pfx.sort()
            self.append(pfx)

        self.sort()



class Event:
    def __init__(self):
        self._listeners = []

    def add_listener(self, listener: callable):
        assert callable(listener), "The given listener must be callable"
        self._listeners.append(listener)

    def remove_listener(self, listener):
        if listener in self._listeners:
            self._listeners.remove(listener)

    def __call__(self, *args, **kwargs):
        for listener in self._listeners:
            listener(*args, **kwargs)



import logging
import os
import time
from datetime import datetime
from glob import glob
from zipfile import ZipFile, ZIP_LZMA

from grapejuice_common import variables

LOG = logging.getLogger(__name__)
N_KEEP_LOGS = 10


def log_files():
    return glob(os.path.join(variables.logging_directory(), "*.log"))


def archive_directory():
    return os.path.join(variables.logging_directory(), "archive")


def archive_files():
    return glob(os.path.join(archive_directory(), "*.zip"))


def archive_logs(old_log_files):
    datetime_now = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    os.makedirs(archive_directory(), exist_ok=True)
    archive_path = os.path.join(archive_directory(), f"{datetime_now}.zip")

    LOG.info(f"Writing log archive {archive_path}")

    with ZipFile(archive_path, "w", ZIP_LZMA) as zf:
        for file in old_log_files:
            with open(file, "rb") as fp:
                zf.writestr(os.path.basename(file), fp.read(), ZIP_LZMA)

    for file in old_log_files:
        LOG.debug(f"Removing log file {file}")
        os.remove(file)


def can_delete_archive(file):
    stat = os.stat(file)
    time_delta = int(time.time()) - int(stat.st_ctime)

    # Older than a week
    return time_delta > 604800


def vacuum_logs():
    files = log_files()

    if len(files) >= 50:
        files.sort(key=lambda f: os.stat(f).st_ctime)
        old_files = files[:-N_KEEP_LOGS]
        archive_logs(old_files)

    for file in filter(can_delete_archive, archive_files()):
        LOG.info(f"Removing log archive {file}")
        os.remove(file)



import os
from typing import Union

from gi.repository import Gtk

from grapejuice_common import robloxctrl
from grapejuice_common import variables
from grapejuice_common.fast_flags import FastFlagList, FastFlag
from grapejuice_common.gtk.GtkPaginator import GtkPaginator
from grapejuice_common.gtk.gtk_stuff import WindowBase
from grapejuice_common.paginator import Paginator


class WidgetStuff:
    def __init__(self, widget, get_value, set_value):
        self.widget = widget
        self.get_value = get_value
        self.set_value = set_value
        self.icon_changes: Union[None, Gtk.Image] = None
        self.reset_button: Union[None, Gtk.Button] = None


def flag_to_widget(flag: FastFlag, on_changed: callable = None) -> Union[None, WidgetStuff]:
    widget = None

    if flag.is_a(bool):
        widget = Gtk.Switch()
        widget.set_active(flag.value)
        widget.set_vexpand(False)
        widget.set_vexpand_set(True)

        def get_value():
            return widget.get_active()

        def set_value(v):
            widget.set_active(v)

        if on_changed is not None:
            widget.connect("state-set", on_changed)

    elif flag.is_a(str):
        widget = Gtk.Entry()
        widget.set_text(flag.value)
        widget.set_hexpand(True)
        widget.set_hexpand_set(True)

        def get_value():
            return widget.get_text()

        def set_value(v):
            widget.set_text(str(v))

        if on_changed is not None:
            widget.connect("changed", on_changed)

    elif flag.is_a(int):
        adjustment = Gtk.Adjustment()
        adjustment.set_step_increment(1.0)
        adjustment.set_upper(2147483647)
        adjustment.set_value(flag.value)

        widget = Gtk.SpinButton()
        widget.set_adjustment(adjustment)
        widget.set_value(flag.value)

        def get_value():
            return int(adjustment.get_value())

        def set_value(v):
            adjustment.set_value(int(v))

        if on_changed is not None:
            adjustment.connect("value-changed", on_changed)

    else:
        return None

    return WidgetStuff(widget, get_value, set_value)


class FastFlagEditor(WindowBase):
    def __init__(self):
        super().__init__(variables.fast_flag_editor_glade(), self)

        self._fast_flags = FastFlagList().import_file(variables.wine_roblox_studio_app_settings())

        client_settings_path = robloxctrl.locate_client_app_settings()
        if client_settings_path is not None and os.path.exists(client_settings_path):
            self._fast_flags.overlay_flags(FastFlagList().import_file(client_settings_path))

        self._paginator = Paginator(self._fast_flags, 50)
        self._gtk_paginator = GtkPaginator(self._paginator)
        self.gtk_pager_box.add(self._gtk_paginator.get_root_widget())

        self._flag_refs = dict()
        self._rows = dict()
        self._displayed_rows = list()

        self._populate()
        self._paginator.paged.add_listener(self._populate)

        self.__unsaved_changes = False

    def _populate(self):
        gtk_list = self.gtk_fast_flag_list

        for row in self._displayed_rows:
            gtk_list.remove(row)

        self._displayed_rows = list()

        for flag in self._paginator.page:
            if flag in self._rows:
                row = self._rows[flag]

            else:
                row = self.new_row(flag)
                self._rows[flag] = row

            gtk_list.add(row)
            self._displayed_rows.append(row)

        self.fast_flag_scroll.get_vadjustment().set_value(0)
        gtk_list.show_all()

        self._update_change_icons()

    @property
    def _unsaved_changes(self):
        return self.__unsaved_changes

    @_unsaved_changes.setter
    def _unsaved_changes(self, v):
        self.__unsaved_changes = v

        if self.__unsaved_changes:
            self.gtk_header.set_subtitle("Unsaved changes!")

        else:
            self.gtk_header.set_subtitle("")

    @property
    def window(self):
        return self.builder.get_object("fast_flag_editor")

    @property
    def gtk_fast_flag_list(self):
        return self.builder.get_object("fast_flag_list")

    @property
    def gtk_pager_box(self):
        return self.builder.get_object("paginator_box")

    @property
    def gtk_header(self):
        return self.builder.get_object("fast_flag_editor_header")

    @property
    def fast_flag_scroll(self):
        return self.builder.get_object("fast_flag_scroll")

    def new_row(self, flag: FastFlag):
        builder = self._create_builder()

        row = builder.get_object("fast_flag_row")

        name_label = builder.get_object("fflag_name_label")
        name_label.set_text(flag.name)

        widgets = builder.get_object("fast_flag_widgets")
        icon_changes = builder.get_object("icon_flag_changes")
        reset_button = builder.get_object("fflag_reset_button")

        stuff = None

        def reset_flag(*_):
            flag.reset()
            stuff.set_value(flag.value)
            reset_button.hide()

        reset_button.connect("clicked", reset_flag)

        def on_widget_changed(*_):
            flag.value = stuff.get_value()
            self._update_change_icons()

            if flag.has_changed and not flag.is_a(bool):
                reset_button.show()

            else:
                reset_button.hide()

            self._unsaved_changes = True

        stuff = flag_to_widget(flag, on_widget_changed)
        stuff.icon_changes = icon_changes
        stuff.reset_button = reset_button

        if stuff and stuff.widget is not None:
            self._flag_refs[flag] = stuff
            widgets.add(stuff.widget)

        wrapper = Gtk.ListBoxRow()
        wrapper.add(row)

        return wrapper

    def _input_values_to_flags(self):
        for flag, ref in self._flag_refs.items():
            flag.value = ref.get_value()

    def _flags_to_inputs(self):
        self._fast_flags.sort()

        for flag in filter(lambda f: f in self._flag_refs, self._fast_flags):
            ref = self._flag_refs[flag]
            ref.set_value(flag.value)

        self._paginator.paged()

    def _update_change_icons(self):
        for flag, ref in self._flag_refs.items():
            if flag.has_changed:
                ref.icon_changes.show()
                if not flag.is_a(bool):
                    ref.reset_button.show()

            else:
                ref.icon_changes.hide()
                ref.reset_button.hide()

            if flag.is_a(bool):
                ref.reset_button.hide()

    def save_flags_to_studio(self, *_):
        self._input_values_to_flags()
        changed_flags = self._fast_flags.get_changed_flags()
        save_path = robloxctrl.locate_client_app_settings()

        changed_flags.export_to_file(save_path)

        self._unsaved_changes = False

    def on_search_changed(self, search_entry):
        query = search_entry.get_text().lower()

        if query:
            def filter_function(flags_list):
                return filter(lambda flag: query in flag.name.lower(), flags_list)

            self._paginator.filter_function = filter_function

        else:
            self._paginator.filter_function = None

    def reset_all_flags(self, *_):
        self._fast_flags.reset_all_flags()
        self._flags_to_inputs()
        self._unsaved_changes = False

    def delete_user_flags(self, *_):
        self.reset_all_flags()

        path = robloxctrl.locate_client_app_settings()
        if os.path.exists(path):
            os.remove(path)
